from asyncio import run
from functools import wraps
from re import findall

from click import command, echo, option

from proxygrab import get_proxy

# Constant Variables
proxy_types = ("http", "https", "socks4", "socks5")
fetch_methods = ("all", "api", "scrapper")

# List to str with ', ' as seperator - for mehods
list_methods = lambda: ", ".join(fetch_methods)

# List to str with ', ' as seperator - for proxy types
list_ptypes = lambda: ", ".join(proxy_types)


with open("proxygrab/__init__.py", encoding="utf-8") as f:
    version_pip = findall(r"__version__ = \"(.+)\"", f.read())[0]

with open("proxygrab/__init__.py", encoding="utf-8") as f:
    copyright_pip = findall(r"__copyright__ = \"(.+)\"", f.read())[0]

# For making it run in async
def coro(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        return run(f(*args, **kwargs))

    return wrapper


@command(
    context_settings=dict(help_option_names=["-h", "--help"]),
    options_metavar="<options>",
)
@option(
    "--method",
    "-m",
    default="all",
    type=str,
    help=f"Method to get proxies form, available: {list_methods()}",
    metavar="<method>",
    show_default=True,
)
@option(
    "--type",
    "-t",
    type=str,
    help=f"Will get the specific type of proxies, only 4 types are availabe right now: {list_ptypes()}",
    metavar="<proxy type>",
)
@option(
    "--outfile",
    "-o",
    type=str,
    help="Will save with specified filename.",
    metavar="<filename>",
)
@option("--save", "-s", is_flag=True, help="Will save proxies to file.")
@option("--version", "-V", is_flag=True, help="Show version of proxygrab.")
@option(
    "--count",
    "-n",
    type=int,
    default=0,
    help="Number of Proxies; 0 means all",
    metavar="<num>",
    show_default=True,
)
@coro
async def clicmd(save, version, type: str, outfile: str, count: int, method: str):
    """
    This a Command Line Utility from ProxyGrab which can be used to get proxies straight in your terminal or to save them to a file.
    """

    # Return the current version
    if version:
        return echo(
            (f"\nProxyGrab Version: {version_pip}\nCopyright: {copyright_pip}\n"),
        )

    # Initially Check if user has provided the proxy type and it is in the supported formats!
    if not type:
        return echo("Check help by proxygrab --help")

    if type not in proxy_types:
        return echo(f"Only following types are supported: {list_ptypes()}")

    type = type.lower()  # Convert proxytype text to lower
    echo("Fetching proxies...")

    # Return if the method provided by user is not available
    if method not in fetch_methods:
        return echo(f"Only following methods are supported: {list_methods()}")

    # Use function to get proxies!
    # Default method is 'all'
    proxies = await get_proxy(type, method)

    # If user has defined the proxies count, scrap them only
    if count != 0:
        if count <= len(proxies):
            proxies = proxies[0:count]

    # If --save flag is not used, print proxies to terminal
    if not save:
        echo(proxies)
        return echo(f"Printed {len(proxies)} {type} proxies to Terminal as list.")

    # If no filename defined, use the default one
    if not outfile:
        filename = f"{type}_proxygrab.txt"

    # Write to file
    with open(filename, "w+") as f:
        for proxy in proxies:
            f.write(str(proxy + "\n"))
        f.close()

    echo(f"Saved to {filename} ^_^")
    echo(f"Number of proxies: {len(proxies)}")

    return



from .._utils import AioHttp


async def proxyscrape(ptype: str):
    """Get Proxies from Proxyscrape website."""
    url = f"https://api.proxyscrape.com/?request=displayproxies&proxytype={ptype}"
    text, rr = await AioHttp.get_text(url)
    if rr.status == 200:
        px = text.split("\r\n")
        proxies = px[1:-2]
        return True, proxies
    return False, "An error occured!\nResponse not equal to 200"



from json import loads

from .._utils import AioHttp


async def get_github_proxies(p_type: str):
    lst = []
    lst += await fate0_proxy(p_type)
    lst += await shiftytr_proxy(p_type)
    lst += await speedx_proxy(p_type)
    return lst


async def fate0_proxy(p_type: str):
    url = "https://raw.githubusercontent.com/fate0/proxylist/master/proxy.list"
    res, _ = await AioHttp.get_text(url)
    tmp_lst = [loads(pxy) for pxy in res.split("\n")[:-2]]
    try:
        p_list = [
            f"{proxy['host']}:{proxy['port']}"
            for proxy in tmp_lst
            if proxy["type"] == p_type
        ]
    except KeyError:
        p_list = []
    return p_list


async def shiftytr_proxy(p_type: str):
    proxy_map = {
        "http": "https://raw.githubusercontent.com/ShiftyTR/Proxy-List/master/http.txt",
        "https": "https://raw.githubusercontent.com/ShiftyTR/Proxy-List/master/https.txt",
        "socks4": "https://raw.githubusercontent.com/ShiftyTR/Proxy-List/master/socks4.txt",
        "socks5": "https://raw.githubusercontent.com/ShiftyTR/Proxy-List/master/socks5.txt",
    }

    res, _ = await AioHttp.get_text(proxy_map[p_type])
    p_list = res.split("\n")
    return p_list


async def speedx_proxy(p_type: str):
    proxy_map = {
        "http": "https://raw.githubusercontent.com/TheSpeedX/PROXY-List/master/http.txt",
        "socks4": "https://raw.githubusercontent.com/TheSpeedX/PROXY-List/master/socks4.txt",
        "socks5": "https://raw.githubusercontent.com/TheSpeedX/PROXY-List/master/socks5.txt",
    }

    # As this one does not have 'https' proxy
    if p_type not in proxy_map.keys():
        return []

    res, _ = await AioHttp.get_text(proxy_map[p_type])
    p_list = res.split("\n")
    return p_list



from bs4 import BeautifulSoup

from .._utils import AioHttp


async def get_free_proxy_list_proxies(ptype: str):
    url = "http://www.free-proxy-list.net"
    response, _ = await AioHttp.get_text(url)

    try:
        soup = BeautifulSoup(response.content, "lxml")
        table = soup.find("table", {"id": "proxylisttable"})
        proxies = []

        for row in table.find("tbody").find_all("tr"):
            data = list(map(lambda x: x.text, row.find_all("td")))
            host = data[0]
            port = data[1]
            if data[6].lower() == "yes":
                version = "https"
            else:
                version = "http"

            if version == ptype:
                proxies.append(f"{host}:{port}")

        return proxies
    except Exception:
        return []


# HTTP and HTTPS Proxies Scrapper
async def get_anonymous_proxiesptype(ptype: str):
    url = "https://free-proxy-list.net/anonymous-proxy.html"
    response, _ = await AioHttp.get_text(url)

    try:
        soup = BeautifulSoup(response.content, "lxml")
        table = soup.find("table", {"id": "proxylisttable"})
        proxies = []

        for row in table.find("tbody").find_all("tr"):
            data = list(map(lambda x: x.text, row.find_all("td")))
            host = data[0]
            port = data[1]
            if data[6].lower() == "yes":
                version = "https"
            else:
                version = "http"

            if version == ptype:
                proxies.append(f"{host}:{port}")

        return proxies
    except Exception:
        return []


async def get_uk_proxies(ptype: str):
    url = "https://free-proxy-list.net/uk-proxy.html"
    response, _ = await AioHttp.get_text(url)

    try:
        soup = BeautifulSoup(response.content, "lxml")
        table = soup.find("table", {"id": "proxylisttable"})
        proxies = []

        for row in table.find("tbody").find_all("tr"):
            data = list(map(lambda x: x.text, row.find_all("td")))
            host = data[0]
            port = data[1]
            if data[6].lower() == "yes":
                version = "https"
            else:
                version = "http"

            if version == ptype:
                proxies.append(f"{host}:{port}")

        return proxies
    except Exception:
        return []



"""Various Scrapper used to scrap proxies from different websites.
You can find a list of scrappers in the readme file"""

from .free_proxy_list import (
    get_anonymous_proxiesptype,
    get_free_proxy_list_proxies,
    get_uk_proxies,
)
from .github import get_github_proxies
from .socks_proxy import get_socks_proxies
from .sslproxies import get_ssl_proxies
from .us_proxy import get_us_proxies


async def grab_proxies(ptype: str):
    lst = []
    if ptype in ("http", "https"):
        if ptype == "https":
            lst += await get_ssl_proxies()
        lst += await get_anonymous_proxiesptype(ptype)
        lst += await get_free_proxy_list_proxies(ptype)
        lst += await get_us_proxies(ptype)
        lst += await get_uk_proxies(ptype)
    else:
        lst += await get_socks_proxies(ptype)

    # Github Scrapper
    lst += await get_github_proxies(ptype)

    return lst



"""initialise api sub-package."""

from .proxylist import proxylist
from .proxyscrape import proxyscrape


async def get_api_proxies(p_type: str):
    lst = []
    lst += (await proxylist(p_type))[1]
    lst += (await proxyscrape(p_type))[1]
    return lst



from .._utils import AioHttp


async def proxylist(ptype: str):
    """Get Proxies from proxy-list.download"""
    url = f"https://www.proxy-list.download/api/v1/get?type={ptype}"
    text, rr = await AioHttp.get_text(url)
    if rr.status == 200:
        px = text.split("\r\n")
        proxies = px[1:-2]
        return True, proxies
    return False, "An error occured!\nResponse not equal  200"



class ProxyGrabError(Exception):
    """General class for handling exceptions."""

    pass


class MethodNotFound(ProxyGrabError):
    """Exception raised for errors in the input method.

    Attributes:
        method -- input method which caused the error
        message -- explanation of the error
    """

    def __init__(
        self,
        method,
        message="Method to grab proxies not in ('api', 'scrappper', 'all')",
    ):
        self.method = method
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        return f"{self.method} -> {self.message}"


class ProxyTypeNotFound(ProxyGrabError):
    """Exception raised for errors in the input proxy type.

    Attributes:
        ptype -- input proxy type which caused the error
        message -- explanation of the error
    """

    def __init__(
        self,
        ptype,
        message="Proxy Type not in ('http', 'https', 'socks4', 'socks5')",
    ):
        self.ptype = ptype
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        return f"{self.ptype} -> {self.message}"


class UnknownError(ProxyGrabError):
    """Exception raised when error is unknown."""

    def __init__(
        self,
        causes=(
            "Some causes of error may be:\n"
            "1. Maybe check you internet connection?\n"
            "2. No Proxies found!\n"
            "3. Maybe your IP is Temporarily Banned!"
        ),
        message="Unknown Error!",
    ):
        self.causes = causes
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        return f"{self.message}\n -> {self.causes}"



from aiohttp import ClientSession


class AioHttp:
    """class for helping get the data from url using aiohttp."""

    @staticmethod
    async def get_json(link: str, **kwargs):
        """Get JSON data from the provided link."""
        async with ClientSession() as sess:
            async with sess.get(link, **kwargs) as resp:
                return await resp.json(), resp

    @staticmethod
    async def get_text(link: str, **kwargs):
        """Get Text data from the provided link."""
        async with ClientSession() as sess:
            async with sess.get(link, **kwargs) as resp:
                return await resp.text(), resp

    @staticmethod
    async def get_raw(link: str, **kwargs):
        """Get RAW data from the provided link."""
        async with ClientSession() as sess:
            async with sess.get(link, **kwargs) as resp:
                return await resp.read(), resp



__version__ = "1.0.2"
__copyright__ = "Divkix <https://github.com/Divkix>"

# Methods to get proxies!
from .package.main import (
    get_all_proxies,
    get_http,
    get_https,
    get_proxy,
    get_socks4,
    get_socks5,
    save_all_proxies,
    save_http,
    save_https,
    save_proxy,
    save_socks4,
    save_socks5,
)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['proxygrab',
 'proxygrab.package',
 'proxygrab.package.api',
 'proxygrab.package.scrappers']

package_data = \
{'': ['*']}

install_requires = \
['aiodns==3.0.0',
 'aiohttp==3.7.4.post0',
 'beautifulsoup4==4.9.3',
 'cchardet==2.1.7',
 'click==8.0.1',
 'lxml==4.6.3']

entry_points = \
{'console_scripts': ['proxygrab = proxygrab.cmdline:clicmd']}

setup_kwargs = {
    'name': 'proxygrab',
    'version': '1.0.2',
    'description': 'Asynchronous Library to scrap proxies for my web scrapping applications and other testing purposes',
    'long_description': '# ProxyGrab\n\n<p align="center">\n<a href="https://proxygrab.netlify.com"><img src="https://raw.githubusercontent.com/Divkix/ProxyGrab/master/docs/img/name.png"></a>\n<i>Software to scrap proxies for my web scrapping and other testing purposes.</i></br></br>\n<a href="https://pypi.org/project/ProxyGrab/"><img src="https://img.shields.io/pypi/v/ProxyGrab" alt="PyPI"></a>\n<a href="https://github.com/Divkix/ProxyGrab/actions"><img src="https://github.com/Divkix/ProxyGrab/workflows/CI%20%28pip%29/badge.svg" alt="CI (pip)"></a>\n<a href="https://www.codacy.com/gh/Divkix/ProxyGrab/dashboard?utm_source=github.com&amp;utm_medium=referral&amp;utm_content=Divkix/ProxyGrab&amp;utm_campaign=Badge_Grade"><img src="https://app.codacy.com/project/badge/Grade/b5b68ed7f04c4f639bef56df0668d289"/></a>\n<a href="https://pypi.org/project/proxygrab/"><img src="https://img.shields.io/pypi/pyversions/ProxyGrab.svg" alt="Supported Python Versions"></a>\n<a href="https://pepy.tech/project/ProxyGrab"><img src="https://pepy.tech/badge/ProxyGrab" alt="Downloads"></a>\n<a href="https://proxygrab.netlify.com"><img src="https://api.netlify.com/api/v1/badges/07f64c3d-03c2-48e5-b947-0c75d38ff8ec/deploy-status" alt="Documentation"></a>\n</p>\n\nI made this software to scrap proxies for my web scrapping and other testing purposes. This program just uses [aiohttp](https://pypi.org/project/aiohttp/) to get the response from API and return the proxies, also it can scrape proxies from a few sites so that it can be used without using the API.\n\n<i><b>NOTE:</b> This library isn\'t designed for production use. It\'s advised to use your own proxies or purchase a service which provides an API. These are merely free ones that are retrieved from sites and should only be used for development or testing purposes.</br>\n\n\n## Installation\n\nThe latest version of proxygrab is available via `pip`:\n\n```shell\npip install --upgrade proxygrab\n```\n\nAlso, you can download the source code and install using:\n\n```shell\npython setup.py install\n```\n\n## Provided Proxies\n\n<table style="width:100%">\n  <tr>\n    <th>Provider</th>\n    <th>Proxy Types avaiable</th>\n    <th>Url</th>\n  </tr>\n  <tr>\n    <td>Proxyscrape</td>\n    <td>http, https, socks4, socks5</td>\n    <td>https://proxyscrape.com/</td>\n  </tr>\n  <tr>\n    <td>Proxy-List</td>\n    <td>http, https, socks4, socks5</td>\n    <td>https://www.proxy-list.download/</td>\n  </tr>\n  <tr>\n    <td>SSL Proxies</td>\n    <td>https</td>\n    <td>https://www.sslproxies.org/</td>\n  </tr>\n  <tr>\n    <td>Free Proxy List</td>\n    <td>http, https</td>\n    <td>https://free-proxy-list.net/</td>\n  </tr>\n  <tr>\n    <td>US Proxies</td>\n    <td>http, https</td>\n    <td>https://www.us-proxy.org/</td>\n  </tr>\n  <tr>\n    <td>Socks Proxy</td>\n    <td>socks4, socks5</td>\n    <td>https://www.socks-proxy.net/</td>\n  </tr>\n</table>\n\n## Documentation\n\nhttps://proxygrab.netlify.com\n\n## Contribuiting\n\nWanna help and improve this project?\n\nMake sure to follow these before opening a PR:\n\n- Make sure your PR passes the test and is formatted according to pre-commit.\n- Make sure the package is working without any issues!\n\n## License\n\nThis project is licensed under the MIT License - see the [LICENSE](LICENSE) file for details\n\n## Thanks to\n\n- [@JaredLGillespie](https://github.com/JaredLGillespie) for his [proxyscrape library](https://github.com/JaredLGillespie/proxyscrape) from which I took scrappers!\n- Proxy Providers mentioned above\n',
    'author': 'Divkix',
    'author_email': 'techdroidroot@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Divkix/ProxyGrab',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6.2,<4.0.0',
}


setup(**setup_kwargs)



"""Main script which compiles all the functions from more different scripts."""

from json import dump

from .api import get_api_proxies
from .errors import MethodNotFound, ProxyTypeNotFound
from .scrappers import grab_proxies

# Constants Start
method_types = ("all", "scrapper", "api")
proxy_types = ("http", "https", "socks4", "socks5")
# Constants End


async def clean(mylist: list):
    """Clean Duplicate proxies from list by first converting list to dictionary and then
    extracting keys from it, as keys have unique value, there won't be any duplicates."""
    return list(set(mylist))


async def get_proxies_func(ptype: str, method: str):
    """Function to get proxies."""
    method = method.lower()  # Convert method name to lowercase
    ptype = ptype.lower()  # Convert proxy name to lowercase

    if method == "all":  # All Method
        l1 = await get_api_proxies(ptype)  # Get proxies from Proxyscrape free API
        l2 = await grab_proxies(ptype)  # Get proxies from scrapper
        all_proxies = l1 + l2

    elif method == "api":  # API Method
        l1 = await get_api_proxies(ptype)  # Get proxies from Proxyscrape free API
        l2 = await grab_proxies(ptype)  # Get proxies from scrapper
        all_proxies = l1 + l2

    elif method == "scrapper":  # Scrapper Method
        all_proxies = await grab_proxies(ptype)

    else:
        # Raise Exception if method is not in ('api', 'scrappper', 'all')
        raise MethodNotFound(method)

    # Clean proxies so that duplicates are removed from list!
    return await clean(all_proxies)


# For HTTP
async def get_http(method: str = "all"):
    """Get http proxies from get_proxies_func() function."""
    return await get_proxies_func("http", method)


async def save_http(filename: str = "http_proxygrab.txt", method: str = "all"):
    """Save http proxies from get_proxies_func() function."""
    proxies = await get_proxies_func("http", method)
    with open(filename, "w+") as f:
        f.write("\n".join(proxies))
    return filename


# For HTTPS
async def get_https(method: str = "all"):
    """Get https proxies from get_proxies_func() function."""
    return await get_proxies_func("https", method)


async def save_https(filename: str = "https_proxygrab.txt", method: str = "all"):
    """Save https proxies from get_proxies_func() function."""
    proxies = await get_proxies_func("https", method)
    with open(filename, "w+") as f:
        f.write("\n".join(proxies))
    return filename


# For SOCKS4
async def get_socks4(method: str = "all"):
    """Get socks4 proxies from get_proxies_func() function."""
    return await get_proxies_func("socks4", method)


async def save_socks4(filename: str = "socks4_proxygrab.txt", method: str = "all"):
    """Save socks4 proxies from get_proxies_func() function."""
    proxies = await get_proxies_func("socks4", method)
    with open(filename, "w+") as f:
        f.write("\n".join(proxies))
    return filename


# For SOCKS5
async def get_socks5(method: str = "all"):
    """Get socks5 proxies from get_proxies_func() function."""
    return await get_proxies_func("socks5", method)


async def save_socks5(filename: str = "socks5_proxygrab.txt", method: str = "all"):
    """Save socks5 proxies from get_proxies_func() function."""
    proxies = await get_proxies_func("socks5", method)
    with open(filename, "w+") as f:
        f.write("\n".join(proxies))
    return filename


# For general - user defines proxy type
async def get_proxy(ptype, method: str = "all"):
    """Get the type of proxies we define using method from get_proxies_func() function."""
    if ptype not in proxy_types:
        raise ProxyTypeNotFound(ptype)
    return await get_proxies_func(ptype, method)


async def save_proxy(ptype, method: str = "all"):
    """Save the type of proxies we define using method from get_proxies_func() function."""
    if ptype not in proxy_types:
        raise ProxyTypeNotFound(ptype)
    proxies = await get_proxies_func(ptype, method)
    filename = f"{ptype}_proxygrab.txt"
    with open(filename, "w+") as f:
        f.write("\n".join(proxies))
    return filename


# For all proxy_types
# These 2 fucntions saves the proxies in dictionary in file
async def get_all_proxies(method: str = "all"):
    pdict = {i.upper(): await get_proxy(i, method) for i in proxy_types}
    return pdict


# save all proxies
async def save_all_proxies(
    filename: str = "all_proxies_proxygrab.txt",
    method: str = "all",
):
    pdict = {i.upper(): await get_proxy(i, method) for i in proxy_types}
    dump(pdict, filename)
    return filename



from bs4 import BeautifulSoup

from .._utils import AioHttp


# Only HTTPS Proxies
async def get_ssl_proxies():
    url = "https://www.sslproxies.org/"
    response, _ = await AioHttp.get_text(url)

    try:
        soup = BeautifulSoup(response.content, "lxml")
        table = soup.find("table", {"id": "proxylisttable"})
        proxies = []

        for row in table.find("tbody").find_all("tr"):
            data = list(map(lambda x: x.text, row.find_all("td")))
            host = data[0]
            port = data[1]
            proxies.append(f"{host}:{port}")

        return proxies
    except Exception:
        return []
